"""
SFCTOOLS 
========

Thomas Baldauf
thomas.baldauf@dlr.de 
November 2021, German Aerospace Center

Licensed under MIT license 

    1. CORE FEATURES
    2. DATA STRUCTURES
    3. AUTOMATION AND PARAMETER-IO
    4. BOTTOM-UP FEATURES 
    5. MISC

"""


"""
CORE FEATURES 
"""
from .core.agent import Agent, SingletonAgent
from .core.agent import block_on_bankrupt
from .core.singleton import Singleton
from .core.clock import Clock 
from .core.settings import Settings
from .core.world import World 
from .core.flow_matrix import FlowMatrix, Accounts 

"""
DATA STRUCTURES
"""
from .datastructs.balance import BalanceSheet, BalanceEntry
from .datastructs.income_statement import IncomeStatement,ICSEntry
from .datastructs.cash_flow_statement import CashFlowStatement
from .datastructs.bank_order_book import BankOrderBook
from .datastructs.inventory import Inventory
from .datastructs.market_registry import MarketRegistry
from .datastructs.worker_registry import WorkerRegistry
from .datastructs.signalslot import Signal,Slot,SignalSlot

"""
AUTOMATION AND PARAMETER-IO
""" 
from .automation.runner import ModelRunner 
from .automation.calibration import CalibrationRoutine, DistanceMeasure

"""
BOTTOM-UP FEATURES
"""
from .bottomup.stock_manager import StockManager
from .bottomup.matching import MarketMatching
from .bottomup.productiontree import ProductionTree

"""
MISCELLANEOUS
"""
from .misc.timeseries import stretch_pandas, stretch_datetime, convert_quarterly_to_datetime, stretch_to_length, convert_numeric
from .misc.mpl_plotting import matplotlib_barplot,matplotlib_lineplot, plot_sankey
from .misc.reporting_sheet import ReportingSheet, DistributionReport, IndicatorReport




__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod

from collections import defaultdict
import pandas as pd
import warnings
from .singleton import Singleton
from enum import Enum
from .world import World
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import networkx as nx
from ..misc.mpl_plotting import plot_sankey

class Accounts(Enum):
    """
    Enum for account type in flow matrix. Current and capital account is allowed
    """
    KA = 0 # capital account
    CA = 1 # current account


class FlowMatrix(Singleton):
    """
    The flow matrix takes care of the flows between agents and the stock-flow consistency. It has the following structure:
    

+--------------------------------------------------------------------------------------+
| Flow Matrix                                                                          |
+===============+=========================+==================================+=========+
|               |     Agent 1             |           Agent 2                |  TOTAL  |
+---------------+-----------+-------------+----------------+-----------------+---------+
|               |     CA    | KA          |           CA   | KA              |         |
+---------------+-----------+-------------+----------------+-----------------+---------+
|               |           |             |                |                 |         |
| Flows         |    -x     |             |                |  +x             |   0     |
+---------------+-----------+-------------+----------------+-----------------+---------+
|               |           |             |                |                 |         |
| d(Stocks)     |     +x    |             |                |  -x             |   0     |
+---------------+-----------+-------------+----------------+-----------------+---------+
| TOTAL         |     0     |  0          |            0   |    0            |   0     |
+---------------+-----------+-------------+----------------+-----------------+---------+

    | CA: Current Account, KA: Capital Account.

    As a consistency check, the sum of rows and columns should be zero after the end of each simulation period. 
    For more information, see

- Investopedia Article https://www.investopedia.com/ask/answers/031615/whats-difference-between-current-account-and-capital-account.asp
- A nice introduction by Gasselli (OECD) https://www.oecd.org/naec/new-economic-policymaking/grasselli_OECD_masterclass_2019.pdf
    
    """
    
    reset_count = 0

    def __init__(self):
        # constructor of the flow matrix 

        if hasattr(self, "initialized"):  # only initialize once 
            return
        
        self.initialized = True # initialized flag 
        
        # TODO replace above code with 'if self.do_init:' check? 
        # TODO Keeping it for now because it works...
        
        self._flow_data = {}  # has two keys CA and KA, see reset()...
        self.reset()  # defaultdict will take care of missing keys here
        self.linear_log = [] # linear transaction log
    
    def reset(self):
        """
        reset the data
        """
        
        self._flow_data[Accounts.CA] = defaultdict(lambda: defaultdict(float))
        self._flow_data[Accounts.KA] = defaultdict(lambda: defaultdict(float))
        
        if self.__class__.reset_count > 1:
            
            # warn the user every time this is being reset!
            warnings.warn("FlowMatrix has been reset") 
        
        else:
            self.__class__.reset_count += 1 
        
    def log_flow(self,kind,quantity,agent_from,agent_to,subject,price=None,invert=False):
        """
        Registers a flow at the flow matrix. This is no method for the user in most cases. It is automatically generated in the qattune gui.

        :param kind: tuple (from_account,to_account)
        :param quantity: weight (quantity) of the flow
        :param price: float or None (default). If not None, a price conversion factor is applied. Should not be used in most cases
        :param invert: reverse sign of the transferred quantity? Default False, should not be used in most cases
        :param agent_from: Agent instance (optionally str), sender agent
        :param agent_to: Agent instnace (optionally str), receiver agent

        Example

        .. code-block:: python

            from sfctools import FlowMatrix, Accounts
            CA = Accounts.CA
            KA = Accounts.KA

            flow = (CA,KA) # from account -> to account
            FlowMatrix().log_flow(flow, 42.0, agent1, agent2, subject="my_subject")
            # ...
            
            df = FlowMatrix().to_dataframe(group=True)
            print(FlowMatrix().to_string(group=True))
            FlowMatrix().check_consistency() # ok if no error is raised

        """

        if price is not None:
            Q = quantity * price
        else:
            Q = quantity

        if invert:
            Q = - Q

        from_account = kind[0] 
        to_account = kind[1]

        self._flow_data[from_account][subject][agent_from]-= Q
        self._flow_data[to_account][subject][agent_to] += Q

    def check_consistency(self):
        """
        Checks the consistency of the flow matrix (i.e.)

        NOTE is slow because of pandas dataframe usage. Do not call often, do not call in excessive for-loops
        TODO test this
        """

        df = self.to_dataframe(group=True)

        if df.empty:
            return

        if not df.empty:
            null_sym = "   .-   "

            # avoid numerical errors
            # so round to at max 4 orders of magntude less than max order of magnitude

            df2 = df.replace(null_sym,0.0).astype(float)
            om_max = int(np.ceil(np.log10(df2.to_numpy().max())))
            om_min = int(np.ceil(np.log10(abs(df2.to_numpy().min()))))
            order_magnitude = max(om_max,om_min)
            df2 = df2.round(-order_magnitude + 4)

            if np.array(df2["Total"]).any():
                raise RuntimeError("Inconsistent Row In Flow Sheet: \n%s" % df.to_string())

            df3 = df2.T
            if np.array(df3["Total"]).any():
                raise RuntimeError("Inconsistent Column In Flow Sheet:\n %s" % df.to_string())

    def to_string(self,group=True):
        """
        converts the flow matrix to string representation
        :return : str
        """
        return self.to_dataframe(group=group).to_string()

    def to_dataframe(self,group =True,insert_nullsym=True):
        """
        Converts the data structure to a human-readable dataframe format.
        WARNING this is slow

        :param group: boolean switch (default True), if True it will group the agents of the same class together
        :param insert_nullsym: will insert '.-' symbol instead of zero, default True
        :return: pandas dataframe object
        """

        df_credit  = pd.DataFrame(self._flow_data[Accounts.CA]).T
        df_capital = pd.DataFrame(self._flow_data[Accounts.KA]).T

        df_merge = pd.concat([df_credit, df_capital], axis=1, keys=['CA', 'KA']).swaplevel(0,1,axis=1).sort_index(axis=1)

        df = df_merge.fillna(0.0).sort_index()

        agent_types = World().get_agent_types()

        if not group:

            df.loc["Total"] = df.sum()
            df["Total"] = df.T.sum()
            return df.round(4)

        else:
            """
            Provide an overview with aggregated classes
            """

            data = {}
            renamer = {}
            for a in agent_types:
                my_group = []
                for b in df.columns:
                    # print("b",b,"a",a,b[0])
                    if isinstance(b[0],a):
                        my_group.append(b[0])

                if len(my_group) > 0:
                    my_df = df.loc[:,df.columns.get_level_values(0).isin(my_group)]
                    my_CA = my_df.loc[:,my_df.columns.get_level_values(1).isin({"CA"})].sum(axis=1)
                    my_KA = my_df.loc[:,my_df.columns.get_level_values(1).isin({"KA"})].sum(axis=1)

                    concat =  pd.concat([my_CA, my_KA], axis=1, keys=['CA', 'KA'])

                    data[a] = concat

                if len(my_group) > 1:
                    plural_suffix = ""

                    if len(World().get_agents_of_type(a.__name__))> 1:
                        plural_suffix = "s"

                    renamer[a] = a.__name__ + plural_suffix
                else:
                    renamer[a] = a.__name__

            if data == {}:
                return pd.DataFrame()

            df2 = pd.concat(data, axis=1)

            df2 = df2.rename(columns=renamer).sort_index()

            # df2 = df2.round(4)

            df2.loc["Total"] = df2.sum()
            df2["Total"] = df2.T.sum()

            df2 = df2.round(5) 
            df2 = df2.round(4) # cut one digit to obtain a 'consistently rounded table' 

            
            if insert_nullsym:
                null_sym = "   .-   "
                df2 = df2.replace(0.0,null_sym)

            df2 = df2.reindex(sorted(df2.columns), axis=1)

            cols = list(df2.columns.values) #Make a list of all of the columns in the df
            #print("COLS",cols)
            cols.pop(cols.index(("Total",""))) #Remove Total from list
            df2 = df2[cols+[("Total","")]] #Create new dataframe with columns in the order you want
            
            return df2

    def plot_colored(self,show_plot=True,group=True):
        """
        Plots the flow matrix as a nice colored heat map.
        This will open up a matplotlib window...

        :param show_plot: show the plot as window? default True. If False, figure is returned instead
        :param group: aggregated view?
        :return fig: figure object
        """

        # TODO nicer formatting
        # TODO more plotting options 

        df = self.to_dataframe(insert_nullsym=False,group=group)
        
        fig = plt.figure(figsize=(9,5))
        sns.heatmap(df, annot=True ,cmap='coolwarm',)
        ax = plt.gca()
        ax.set_ylabel('')    
        ax.set_xlabel('')

        plt.tight_layout()
        if show_plot:
            plt.show()  

        return fig 

    def plot_sankey(self,show_values=True, show_plot=True):
        """
        plots a sankey diagram of the flow matrix

        :param show_values: boolean switch to plot the values of the edge weights
        :param show_plot: show the plot as window? default True. If False, figure is returned instead.
        :return fig: figure object
        """
        df = FlowMatrix().to_dataframe(insert_nullsym=False)

        df_CA = df.iloc[:, df.columns.get_level_values(1)=='CA']
        df_CA.columns = df_CA.columns.droplevel(1) # current account

        df_KA = df.iloc[:, df.columns.get_level_values(1)=='KA'] # capital account 
        df_KA.columns = df_KA.columns.droplevel(1) 

        source  = []
        target = []
        types = []
        value = []

        source2  = []
        target2 = []
        types2 = []
        value2 = []

        subjects = {}
        for i,k in enumerate(df.index):
            subjects[str(k)] = i 
        
        for i, df_i in enumerate([df_CA,df_KA]):
                
            for index, row in df_i.iterrows():
                
                for column in df_i.columns:
                    
                    val = float(row[column])
                    if val < 0.0 and index != "Total" and (not index.startswith("Δ")): # is a source
                        source.append(str(column)+ " source")
                        target.append(str(index) )    
                        value.append(-val)
                        types.append(subjects[str(index)])

                    elif val > 0.0 and index != "Total" and (not index.startswith("Δ")): # is a sink
                        source2.append(str(column) + " sink")
                        target2.append(str(index) )    
                        value2.append(val)
                        types2.append(subjects[str(index)])
                        
        my_sankey_source = pd.DataFrame({"from":source,
                                  "to":target,
                                  "color_id":types,
                                  "value":value}).round(2)
        my_sankey_sink =  pd.DataFrame({"to":source2,
                                  "from":target2,
                                  "color_id":types2,
                                  "value":value2}).round(2)

        fig = plot_sankey([my_sankey_source,my_sankey_sink],show_values=show_values,show_plot=show_plot)
        return fig 

        
    @property
    def capital_flow_data(self):
        """'shortcut' property. only returns the data from the capital account"""
        return self._flow_data[Accounts.KA]

    @property
    def current_flow_data(self):
        """'shortcut' property. only returns the dta fromt he current flow account"""
        return self._flow_data[Accounts.CA]



__author__ = "Thomas Baldauf"
__version__ = "Pre-release (michelada)"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__version__ = "0.3"

import numpy as np
from collections import deque

class ShiftLogger(deque):
    """
    shifting logger register data structure.
    (Same as shift register in electronics.)

    [ a , b , c , d , e ] < f
      ^

    inserting a new value f will cause a to be dropped and all values shifted

    [ b , c , d , e, f ] < ...
      ^

    possible operations are:

    - insert() inserts a new value to the logger
    """

    def __init__(self, length=10):
        """"
        Constructor of the shift register,
        :param length: length of the register memory
        """
        super().__init__(maxlen=length)

    def predict(self, method="mean", dt=1):
        """
        predict dt time steps ahead
        :param method: prediction method to use

        * 'mean' is just the mean value
        * 'trend' uses the trend between first and last value

        :param dt:  number of time steps to predict ahead

        :return: the predicted value for t(0) + dt
        """

        if method == "mean":
            return np.mean(self)

        elif method == "trend":
            return self[-1] + (self[-1] - self[0]) / len(self) * dt

        raise NotImplementedError("Cannot find the prediction method %s" % method)



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

import numpy as np
import matplotlib.pyplot as plt
import warnings

"""
This file embraces the following features

IndicatorReport
DistributionReport
ReportingSheet

WARNING BETA, under development, use at own risk!
"""

class IndicatorReport:
    """
    A generic class for logging scalar values ('indicators').
    """

    _instances = {}

    def __init__(self,xlabel,ylabel):
        """

        :param xlabel: x axis label, str
        :param ylabel: y axis label, str
        """
        self.data = []
        self.xlabel = xlabel
        self.ylabel = ylabel

        if not self.ylabel in self.__class__._instances:
            self.__class__._instances[self.ylabel] = self
        else:
            warnings.warn("Tried to add a report with a variable that is already registered. Skipping...")

    @classmethod
    def getitem(cls,key):
        """
        retrieves a certain report from the instances created.

        :param key: the ylabel of the respective report 
        """
        return cls._instances[key]

    def add_data(self,x):
        """ 
        inserts some data into the data structure

        :param x: a scalar value 
        """
        self.data.append(x)

    def plot_data(self,ax):
        """
        plots the data onto a matplotlib axis

        :param ax: a matplotlib axis
        """

        data = self.data
        x = np.arange(len(data)) + 1
        y = data # np.random.rand(len(x))

        if len(data) > 1 and data[1] is not None:
            has_labels = True

        if len(y) < 100: 
            ax.scatter(x, y, color="gray", s=4.0)  # <- manually optimized
        else:
            ax.plot(x, y, "-", color="black") # <- manually optimized

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

        # ax.set_title(self.title)


class DistributionReport:
    """
    A generic class for a logger of a distribution (i.e. )
    """

    def __init__(self,title):
        """
        :param title: title of the report
        """

        self.data = []
        self.title = title

    def add_data(self, x, label=None):
        """
        adds data into the data structure. (stores a sorted version of x).

        :param x: list or numpy array to store
        :param label: some additional tag
        """

        y = np.sort(x.copy())

        # store distribution
        self.data.append((y,label))

    def plot_data(self, ax,colors=None):
        """
        plot the data onto a matplotlib axis
        
        :param ax: a matplotlib axis
        :param colors: list of colors to use (default None)
        """

        colors = ["gray", "indianred", "mediumblue"]

        has_labels = False

        for i,data in enumerate(self.data):
            x = np.arange(len(data[0]))+1
            y = data[0] # np.random.rand(len(x))
            label = data[1] or ""
            if data[1] is not None:
                has_labels = True
            ax.plot(x, y,color=colors[i%len(colors)], alpha=1-0.8*i/len(self.data[0]),label=label)
            ax.scatter(x, y, color=colors[i % len(colors)], s=.5)

        ax.set_ylabel(self.title)
        ax.set_xlabel("Rank")

        if has_labels:
            ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)


class ReportingSheet:
    """
    Reporting sheet is an overview sheet for reporting model results in form of a grid plot.
    """

    def __init__(self,instances=None):
        """
        :param instances: default None, if not None: list of Report instances can be passed here
        """
        if instances is None:
            self.items = []
        else:
            self.items = instances
    
    def add_report(self,report):
        """
        Adds a report item to this sheet.

        :param report: any IndicatorReport or DistributionReport
        """
        self.items.append(report)

    def plot(self, show_plot=True):
        """
        Generates a plot grid using matplotlib

        :param show_plot: show the figure in a window? default True. If False, figure is returned
        :return fig: matplotlib figure
        """
        
        plt.rcParams.update({'font.size': 8})

        # layout is 2 plots per column...
        n_cols = 2
        n_rows = max(1,int(np.ceil(len(self.items)/n_cols)))

        print("n_rows","n_cols",n_rows,n_cols)
        cm = 1 / 2.54  # centimeters in inches
        fig = plt.figure(figsize=(0.5*3.1415*10.5*cm,14.8*cm)) # DIN A6

        try:
            plt.style.use("macro")
        except:
            pass

        for i,item in enumerate(self.items):
            ax = fig.add_subplot(n_rows,n_cols,i+1)
            item.plot_data(ax)

        plt.tight_layout()
        fig.subplots_adjust(hspace=.67)

        if show_plot:
            plt.show()
        
        return fig 

    def to_latex(self):
        """
        Generates latex code NOT YET IMPLEMENTED
        """
        raise NotImplementedError("Not yet implemented. Wait for beta release.")


if __name__ == "__main__":

    rep_sheet = ReportingSheet()

    for i in range(3):
        my_dist = DistributionReport(title="Distribution 1")
        my_dist.add_data(np.random.rand(10),label="t=1")
        my_dist.add_data(np.random.rand(10),label="t=2")
        my_dist.add_data(np.random.rand(10), label="t=3")
        rep_sheet.add_report(my_dist)

    for i in range(3):
        my_ind = IndicatorReport(xlabel="Time", ylabel="Some Indicator")
        my_ind.add_data(np.random.rand(10), label="t=1")
        my_ind.add_data(np.random.rand(10), label="t=2")
        my_ind.add_data(np.random.rand(10), label="t=3")
        rep_sheet.add_report(my_ind)

    rep_sheet.plot()




__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod

import pandas as pd
from ..core.settings import Settings
import warnings
from ..bottomup.stock_manager import StockManager
from ..core.flow_matrix import FlowMatrix
import numpy as np
from collections import defaultdict
from ..misc.mpl_plotting import matplotlib_barplot
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from enum import Enum
from collections import defaultdict

import copy

class BalanceEntry(Enum):
    """
    these are the different entry types 
    """
    ASSETS = 0
    EQUITY = 1
    LIABILITIES = 2
    TOTAL = 3


class BalanceSheet:
    """
    This is the balance sheet of an agent (it could be used as a standalone datastructure, too, in theory). A balance sheet is of the following form

+------------------+-------------------------+
|    ASSETS        | LIABILITIES AND EQUITY  |
+==================+=========================+
|                  |                         |
|   ...            |   Liabilities           |
+------------------+-------------------------+
|                  |   ...                   |
+------------------+-------------------------+
|                  |   Equity                |
+------------------+-------------------------+
|                  |   ...                   |
+------------------+-------------------------+                                                                
|TOT ASSETS     = TOT LIABILITIES AND EQUITY |
+--------------------------------------------+                                                                
    
    more information can be found here https://www.investopedia.com/terms/b/balancesheet.asp for example.

    """

    class Proxy:
        """
        Proxy class for balance sheet. It can be either engaged (consistent and 'active') or disengaged (inconsistent and 'disabled').
        The consistency is automatically checked inside the 'engage()' call when switched from disengaged to engaged.
        """
        def __init__(self, parent):
            self.parent = parent

        def __enter__(self):
            self.parent._engaged = False

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.parent.engage()

    def __init__(self, owner):
        """
        Balance sheet constructor. Creates a new balance sheet
        
        :param owner: the owner agent of this balance sheet. Can be None for standalone purposes, but it is recommended to pass an Agent instance here.
        """

        self.owner = owner # stores owner instance, if any

        default_value = np.nan 
        self._sheet = defaultdict(lambda: {BalanceEntry.ASSETS: default_value,
                                          BalanceEntry.EQUITY: default_value,
                                          BalanceEntry.LIABILITIES: default_value}) # the actual data structure is a three-column table for assets, equity and liabilities
                                        
        self._sheet[BalanceEntry.TOTAL] = {BalanceEntry.ASSETS: default_value,
                               BalanceEntry.EQUITY: default_value,
                               BalanceEntry.LIABILITIES: default_value} # the 'total' row at the bottom

        self._engaged = True  # <- this can deactivate the consistency check temporarily for modification. Manual modification is not allowed (therefore private).
        self._bankrupt = False # <- this stores if the owner went bankrupt. manual modification should not be needed (therefore private).

    @property 
    def raw_data(self):
        """
        get the raw data of this data structure in dictionary format. 
        This will create a copy of the original data, so the user won't operate on the data directly (which is forbidden)
        """
        d = copy.deepcopy(self._sheet) # deep copy because its a nested dict TODO compress the data somehow
        del d[BalanceEntry.TOTAL]
        return d
        
    #def items(self):
    #    return self.raw_data.items()
        
    #def values(self):
    #    return self.raw_data.values()

    #def keys(self):
    #    return self.raw_data.keys() 

    def __repr__(self):
        return "<Balance Sheet of %s>" % self.owner

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key):
        """
        Get an entry of the balance sheet
        :param key: str, name of the asset
        :return: a dict with ASSETS, EQUITY, LIABILITIES entries of this item in this particular agent. NOTE In most cases, at least 2 of the entries will be zero.
        """
        return self._sheet[key]

    @property
    def modify(self):
        """Modification decorator. Temporarily disengages the data structure. This can be used in a with block:
       
        .. code-block:: python
        
            with balance_sheet_1.modify:
                with balance_sheet_2.modify:
                    balance_sheet_1.change_item(...)
                    balance_sheet_2.change_item(...)
        
        """
        return self.Proxy(self)

    def restore_after_bankrupcy(self,verbose=False):
        # helper function to reset bankrupt flag to False again after bankruptcy has been fixed (e.g. replacement of agent by new fictitious entrant)
        if verbose:
            warnings.warn("%s restore after bankrupcy" % self.to_string()) # TODO < test this
        
        self._bankrupt = False

    def disengage(self):
        """
        'unlocks' the balance sheet -> can be modified now without error
        """
        self._engaged = False

    def engage(self):
        """locks the balance sheet -> no longer can be modified
         performs also a cross-check if consistency is maintained
         """

        eps = 1e-6 # tolerance level for cross-check

        a = 0.0 # dummy for sum of entries
        e = 0.0 
        l = 0.0

        # iterate through the rows of the balance sheet and sum up assets, liabilities, equity
        for entry, data in self._sheet.items():

            if entry == BalanceEntry.TOTAL:
                continue

            a_i = data[BalanceEntry.ASSETS]
            e_i = data[BalanceEntry.EQUITY]
            l_i = data[BalanceEntry.LIABILITIES]

            a += a_i
            e += e_i
            l += l_i

            if a_i < -eps or l_i < -eps:

                # raise RuntimeError("...")

                if entry == "Cash":
                    if not self._bankrupt:
                        self._bankrupt = True

                        warnings.warn("filed bankruptcy for %s\n" %self.owner)
                        # print(self.owner.balance_sheet.to_string())

                        self.owner.file_bankruptcy(event="negative cash")

                    return None
                else:
                    s = self.to_string()
                    error = ValueError("%s of %s reached a value below zero after engage() call" % (entry, self.owner) + "\n\n" + s)
                    raise error

        if abs(a - (e+l)) > eps: # check bookkeeping equation for consistency within tolerance level eps

            # almost zero? bookkeeping equation obeyed
            if a !=0 and abs(a - (e+l))/abs(a) < eps:
                
                pass
                # s = "" 
                
                # relative deviation is still small?  -> just warn
                # warnings.warn("balance sheet of %s is numerically instable. Deviation %.10f" % (self.owner, a-(e+l)) + "\n\n" + s)

            else: # not almost zero? bookkeping equation violated

                s = self.to_string()
                error = ValueError("balance sheet of %s is corrupted after cross-check. Deviation %.10f" % (self.owner, a-(e+l)) + "\n\n" + s)
                raise error

        if e < 0 and e < -eps: # negative equity? = 'balance sheet under water'

            if not self._bankrupt:
                self._bankrupt = True

                # warnings.warn("file bankruptcy for %s\n" %self.owner)

                self.owner.file_bankruptcy(event="negative equity")
            
            else: 
                # already bankrupt -> do not trigger bankruptcy a second time... # TODO < think about logic in here
                pass 
                
            return None

    def get_balance(self, key, kind=BalanceEntry.ASSETS) -> float:
        """
        Gets amount of a certain entry for certain key

        :param key: name of entry (e.g. 'Cash' or 'Apples')
        :param kind: preferrably BalanceEntry: Asset, Liabilities or Equity, optionally str 'Assets', 'Equity' or 'Liabilities'

        :return: float, value in balance sheet
        """

        # TODO automatically detect which kind it is if the other two are zero(?) instead of default value being ASSETS

        if isinstance(kind, str):
            mydict = {
                "Assets": BalanceEntry.ASSETS,
                BalanceEntry.ASSETS: BalanceEntry.ASSETS,
                "Equity": BalanceEntry.EQUITY,
                BalanceEntry.EQUITY: BalanceEntry.EQUITY,
                "Liabilities": BalanceEntry.LIABILITIES,
                BalanceEntry.LIABILITIES: BalanceEntry.LIABILITIES,
            }
            which = mydict[kind]
        else:
            which = kind

        return_value = self._sheet[key][which]
        if np.isnan(return_value):
            return 0
        else:
            return return_value

    def to_string(self):
        """
        Coverts the whole balance sheet to a pandas dataframe and subsequently to a string
        
        :return: str, balance sheet string representation
        """
        return "\n\n" + self.to_dataframe().to_string().replace("NaN"," ~ ") + "\n\n"

    def change_item(self, name, which, value, suppress_stock=False):
        """
        Chanes value of an item in the balance sheet. 

        :param name: name of the asset / entry to change
        :param value: value is added to entry. can also be negative
        :param which: BalanceEntry: Asset, Liability, Equity
        """

        """
        if isinstance(which, str):
            mydict = {
                "Assets": BalanceEntry.ASSETS,
                BalanceEntry.ASSETS: BalanceEntry.ASSETS,
                "Equity": BalanceEntry.EQUITY,
                BalanceEntry.EQUITY: BalanceEntry.EQUITY,
                "Liabilities": BalanceEntry.LIABILITIES,
                BalanceEntry.LIABILITIES: BalanceEntry.LIABILITIES,
            }

            which = mydict[which]
        """

        if value == 0:  # security lock if this is a 'ghost transaction'
            return


        if self._engaged:
            raise PermissionError("Cannot change item in engaged balance sheet! This would lead to inconsistencies. Please disengage before calling change_item")

        if np.isnan(self._sheet[name][which]): # first time initialized? -> change nan to zero
            self._sheet[name][which] = 0 
        
        if np.isnan(self[BalanceEntry.TOTAL][which]): # first time initialized? -> change nan to zero
            self._sheet[BalanceEntry.TOTAL][which] = 0 


        self._sheet[name][which] += value
        self._sheet[BalanceEntry.TOTAL][which] += value

        if not suppress_stock:
            if which == BalanceEntry.ASSETS:
                FlowMatrix().capital_flow_data["Δ %s" % name][self.owner] += -value #-value
            if which == BalanceEntry.LIABILITIES:
                FlowMatrix().capital_flow_data["Δ %s" % name][self.owner] += value #value

            if which == BalanceEntry.EQUITY and name == "Cash":
                raise ValueError("Cash cannot be equity")

    """
    TODO insert operations like 'Aktivtausch', 'Passivtausch', 'Bilanzverlängerung'(?)
    TODO < insert code here
    """


    def to_dataframe(self):
        """
        Create a dataframe from this data structure. Warning: this is computationally heavy and should not be used in loops!

        :return: pandas dataframe
        """

        df = pd.DataFrame({"Commodity": [],"Assets": [], "Equity": [],"Liabilities": []}) # construct an empty 'dummy' data frame
        coms = [] # list of commodities

        null_sym = "   .-   " # symbol for zero entries

        # iterate through the balance sheet items and 'fill the table'
        for key, row in self._sheet.items():

            if key != BalanceEntry.TOTAL: # consider all rows except the 'total' entry

                coms.append(key) # commodities

                new_row = pd.DataFrame({"Commodity": [key], 
                                        "Assets": [row[BalanceEntry.ASSETS]],
                                        "Equity": [row[BalanceEntry.EQUITY]],
                                        "Liabilities": [row[BalanceEntry.LIABILITIES]]}) # construct a new row 

                df = df.append(new_row.replace(0.0,null_sym)) # append the new row

        df = df.sort_index() # sort the entries by index

        prices = [1.0 for key in coms] # dummy for prices. NOTE TODO Might change with exchange rates or any other conversion factors (?)
        # for now just assume all values are anyway nominal and therefore no adjustment is required. 

        nominal_row = pd.DataFrame({"Commodity": ["Total"],
                                "Assets": [np.dot(df["Assets"].replace(null_sym,0.0),prices)],
                                "Equity": [np.dot(df["Equity"].replace(null_sym,0.0),prices)],
                                "Liabilities": [np.dot(df["Liabilities"].replace(null_sym,0.0),prices)]})
        df = df.append(nominal_row)

        df = df.set_index("Commodity") # sets index column correctly
        df.index.name = "BALANCE SHEET OF %s" % self.owner # gives the balance sheet an understandable title

        return df 

    def depreciate(self):
        """
        (BETA) this will look up all the required depreciation rates in the
        settings and depreciate the balance sheet.
        Equity will become less, liabilities will stay.
        """

        settings = Settings()
        # read depreciation rates from settings
        for item_name in self._sheet.keys():
            
            depr_rate = settings.config_data["params"][item_name]["depreciation"]
            depr_q = self._sheet[item_name][BalanceEntry.EQUITY] * depr_rate

            warnings.warn("Depreciation is percentage of current stock. Non-linear!")

            if not np.isnan(self._sheet[item_name][BalanceEntry.EQUITY]):
                self._sheet[item_name][BalanceEntry.EQUITY] -= depr_q
            
            if not np.isnan(self._sheet[item_name][BalanceEntry.ASSETS]):
                self._sheet[item_name][BalanceEntry.ASSETS] -= depr_q

    @property
    def leverage(self) -> float:
        """
        computes the financial leverage value (debt-to-equity ratio)
        of this balance sheet

        .. math::
            Leverage = LIABILITIES / EQUITY
        
        :return: float, leverage
        """

        E = self.net_worth
        L = self.total_liabilities

        if E + L == 0:
            return np.inf

        lev = L / (E+L)

        return max(0,lev)

    @property
    def net_worth(self) -> float:
        """net worth = sum of equity
        """
        return self._sheet[BalanceEntry.TOTAL][BalanceEntry.EQUITY] #"Equity"]

    @property
    def total_assets(self):
        """
        total assets = sum of assets column
        """
        return self._sheet[BalanceEntry.TOTAL][BalanceEntry.ASSETS] # "Assets"]

    @property
    def total_liabilities(self):
        """
        total liabilities = sum of all liabilities in liability column
        """
        return self._sheet[BalanceEntry.TOTAL][BalanceEntry.LIABILITIES] # "Liabilities"]


    def plot(self):
        """
        creates a matplotlib plot of the balance sheet
        """

        balance_sheet = self 

        title = "%s" % balance_sheet.owner

        assets = []
        liabilities = []

        colors = []
        greens = ["honeydew", "lightgreen", "seagreen", "darkgreen", "darkslategray"]
        blues = ["lightsteelblue", "lavender", "powderblue", "steelblue", "royalblue", "blue"]
        reds = ["salmon", "indianred", "rosybrown", "orangered", "red", "indianred"]

        greens.reverse()
        blues.reverse()
        reds.reverse()

        legend_patches = []
        blue_patches = []
        green_patches = []
        red_patches = []

        data = {"Assets": [], "Liabilities & Equity": []}
        labels = {}
        i = 0

        assets = []
        lia = []
        equi = []

        for key, val in balance_sheet._sheet.items():

            if key != BalanceEntry.TOTAL:

                if val[BalanceEntry.LIABILITIES] != 0.0:  # is liability

                    # data["Assets"].append(0.0)
                    # data["Liabilities & Equity"].append(val["Liabilities"])
                    assets.append([key, 0.0])
                    lia.append([key, val[BalanceEntry.LIABILITIES]])

                if val[BalanceEntry.EQUITY] != 0.0:  # is Equity

                    # data["Assets"].append(0.0)
                    # data["Liabilities & Equity"].append(val["Equity"])

                    assets.append([key, 0.0])
                    equi.append([key, val[BalanceEntry.EQUITY]])

                if val[BalanceEntry.ASSETS] != 0.0:  # is Asset

                    # data["Assets"].append(val["Assets"])
                    # data["Liabilities & Equity"].append(0.0)

                    assets.append([key, val[BalanceEntry.ASSETS]])
                    lia.append([key, 0.0])

        for a in assets:

            key = a[0]
            val = a[1]

            if val != 0:
                data["Assets"].append(val)
                data["Liabilities & Equity"].append(0.0)

                col = reds.pop()
                colors.append(col) # (col, "black"))
                red_patches.append(mpatches.Patch(color=col, label=key))

                labels[i] = key
                i += 1

        for e in equi:
            key = e[0]
            val = e[1]

            if val != 0:
                data["Assets"].append(0.0)
                data["Liabilities & Equity"].append(val)

                col = blues.pop()
                colors.append(col) # ("black", col))
                blue_patches.append(mpatches.Patch(color=col, label=key))

                labels[i] = key
                i += 1

        for l in lia:

            key = l[0]
            val = l[1]

            if val != 0:
                data["Assets"].append(0.0)
                data["Liabilities & Equity"].append(val)

                col = greens.pop()
                colors.append(col) #("black", col))
                green_patches.append(mpatches.Patch(color=col, label=key))

                labels[i] = key
                i += 1

        legend_patches = red_patches + blue_patches + green_patches

        df = pd.DataFrame(data).T
        df = df.rename(columns=labels).T

        #print("PLOT\n",df)
        #print("colors",colors)

        fig = matplotlib_barplot(df, xlabel="", ylabel="", color=colors, title=title, stacked=True, fmt="{0:.2f}",
                                legend="off", size=(5, 5), show=False)
        plt.tight_layout()

        plt.legend(handles=legend_patches, bbox_to_anchor=(1.25, 1))

        plt.tight_layout()

        plt.show()
        
    
    @classmethod
    def plot_list(cls,list_of_balance_sheets,dt=1,xlabel=None,ylabel=None,title=None,show_liabilities=True):
        """
        Plots a list of balance sheets (e.g. a collected temporal series)

        :param list_of_balance_sheets: a list of BalanceSheet instances
        :param dt: step, (how many values to skip in between)
        :param xlabel: x axis label
        :param ylabel: y axis label 
        :param title: title of the figure
        :param show_liabilities: boolean switch to show passive side of balance sheet as downward-pointing bars (default True)
        """
        
        entries = {}

        for data_i in list_of_balance_sheets:
            
            # print("DATA_I",data_i)
            #print("\n\n --\n\n")

            for k,v in data_i.items():

                if k not in entries:
                    entries[k] = {}

                for k2, v2 in v.items():

                    if k2 not in entries[k]:
                        entries[k][k2] = []

                    if not np.isnan(v2):

                        if k2 == BalanceEntry.ASSETS:
                            entries[k][k2].append(np.abs(v2))
                        else:
                            if show_liabilities:
                                entries[k][k2].append(-np.abs(v2))
                            
        
        #print("ENTRIES",entries)


        plt.figure()

        if xlabel is not None:
            plt.xlabel(xlabel)
        if ylabel is not None:
            plt.ylabel(ylabel)
        if title is not None:
            plt.title(title)

        maxy = 0.01 # np.-1inf
        miny = -0.01 # np.inf

        try:
            plt.style.use("macro")
        except:
            pass

        plt.grid(False)
        # plot upper half

        colors = ["red","maroon","lightcoral","gold","crimson","orange"] # plot colors
        color_idx = 0
        last_vals = None 

        for k,v in entries.items():
            
            for k2,v2 in v.items():

                if  k2 == BalanceEntry.ASSETS:
                    if len(v2) > 0:
                        idx = [int(i) for i in np.arange(0,len(v2)-1,dt)] 
                        # idx2 = [str(Clock().get_i) for i in idx]
                        idx2 = [str(i) for i in idx]

                        print("idx2",idx2)
                        vals = [v2[i] for i in idx]
                        # print("vals",vals)

                        maxy = max(maxy,np.max(vals))
                        miny = min(miny,np.min(vals))
                        
                        plt.bar(idx2,vals,label=str(k),color=colors[color_idx],alpha=0.9,bottom=last_vals)
                        
                        if last_vals is None:
                            last_vals = np.array(vals)
                        else:
                            last_vals += np.array(vals)

                        color_idx += 1 
                        color_idx = color_idx % len(colors)

        
        colors = ["black","blue","navy","dodgerblue","royalblue","slateblue"] # plot colors
        color_idx = 0
        last_vals = None 

        for k,v in entries.items():
            for k2,v2 in v.items():

                if  not k2 == BalanceEntry.ASSETS:
                    
                    if len(v2) > 0:
                        idx = [int(i) for i in np.arange(0,len(v2)-1,dt)] 
                        # idx2 = [str(Clock().get_i) for i in idx]
                        idx2 = [str(i) for i in idx]

                        print("idx2",idx2)
                        vals = [v2[i] for i in idx]
                        # print("vals",vals)

                        maxy = max(maxy,np.max(vals))
                        miny = min(miny,np.min(vals))
                        
                        plt.bar(idx2,vals,label=str(k),color=colors[color_idx],alpha=0.9,bottom=last_vals)
                        
                        if last_vals is None:
                            last_vals = np.array(vals)
                        else:
                            last_vals += np.array(vals)

                        color_idx += 1 
                        color_idx = color_idx % len(colors)
        
        # downward ticks 
        # https://stackoverflow.com/questions/50571287/how-to-create-upside-down-bar-graphs-with-shared-x-axis-with-matplotlib-seabor
        
        plt.xticks(rotation=90)

        #if show_liabilities:
        #    plt.ylim([1.3*miny,1.8*maxy+ .2*max(0,len(entries.keys())-3)])
        #else:
        #    plt.ylim([1.3*miny,1.4*maxy+ .2*max(0,len(entries.keys())-3)])


        ticks =  plt.gca().get_yticks()
        myticks_abs = [(abs(tick)) for tick in ticks]
        myticks = [((tick)) for tick in ticks]
        plt.gca().set_yticks(myticks)
        plt.gca().set_yticklabels(myticks_abs)
        
        plt.legend(loc=(1.04,0)) # "upper left")

        plt.tight_layout()
        plt.show()



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
from networkx.convert_matrix import to_pandas_adjacency,to_pandas_edgelist
from sfctools.src.agent import Agent
from networkx.algorithms.traversal.beamsearch import bfs_beam_edges


class MatrixStruct:
    """
    wrapper for a special networkx graph

    Special data structure for a 'matrix' filled with agents. Will be
    used for aggregate flow matrices. The idea behind this is to have
    a sparse and flexible data structure where agents and their relations can be stored.
    This ensures stock-stock consistency of a macro economic model.
    """

    def __init__(self,handler,name=None):
        """
        set up a new matrixstruct.
        handler will be called when traversing the network for connections
        """

        self.graph = nx.DiGraph()

        self.name = name

        self.handler = handler

    def establish_connection(self, agent1, agent2):
        self.graph.add_edge(agent1, agent2)

    def remove_connection(self, agent1, agent2):
        self.graph.remove_edge(agent1, agent2)

    def plot(self,with_labels=False):
        plt.figure(figsize=(5,5))

        if self.name is None:
            plt.title("My Agent Network")
        else:
            plt.title(self.name)

        nx.draw(self.graph, with_labels=with_labels,node_size=20,node_color='#686f82')
        plt.show()

    def to_dataframe(self, option="adjacency"):
        """
        convert the current graph to a dataframe

        :param option: 'adjacency' or 'edge'
        """

        if option == "adjacency":
            return to_pandas_adjacency(self.graph)
        elif option == "edge":
            return to_pandas_edgelist(self.graph)

    def traverse(self):
        """
        traverses the network and executes the handler for every connection
        """
        for u,v in self.graph.edges:
            u.trigger(self.handler, v)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sfctools',
 'sfctools.automation',
 'sfctools.bottomup',
 'sfctools.core',
 'sfctools.datastructs',
 'sfctools.misc']

package_data = \
{'': ['*']}

install_requires = \
['attrs==19.3.0',
 'cattrs==1.0.0',
 'graphviz>=0.19,<0.20',
 'matplotlib>=3.4,<4.0',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.20.1,<2.0.0',
 'pandas>=1.0,<2.0',
 'poetry>=1.1.6,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'seaborn>=0.11.2,<0.12.0',
 'setuptools>=60.0.0,<61.0.0',
 'wheel>=0.36.2,<0.37.0']

setup_kwargs = {
    'name': 'sfctools',
    'version': '0.4.3',
    'description': 'Framework for stock-flow consistent agent-based modeling, being developed at the German Aerospace Center (DLR) for and in the scientific context of energy systems analysis, however, it is widely applicable in other scientific fields.',
    'long_description': '# sfctools - A toolbox for stock-flow consistent, agent-based models\n\nSfctools is a lightweight and easy-to-use Python framework for agent-based macroeconomic, stock-flow consistent (ABM-SFC) modeling. It concentrates on agents in economics and helps you to construct agents, helps you to manage and document your model parameters, assures stock-flow consistency, and facilitates basic economic data structures (such as the balance sheet).\n\n\n## Installation \n\nIn a terminal of your choice, type: \n\n    pip install sfctools \n\nsee https://pypi.org/project/sfctools/\n\n\n## Usage\n\n```console\nfrom sfctools import Agent,World\nclass MyAgent(Agent):\n    def __init__(self, a):\n        super().__init__(self)\n        self. = a \nmy_agent = MyAgent()\nprint(my_agent)\nprint(World().get_agents_of_type("MyAgent"))\n```\n\n\n| Author Thomas Baldauf, German Aerospace Center (DLR), Curiestr. 4 70563 Stuttgart | thomas.baldauf@dlr.de | version: 0.4 (Beta) | date: February 2022\n',
    'author': 'Thomas Baldauf, Benjamin Fuchs',
    'author_email': 'thomas.baldauf@dlr.de, benjamin.fuchs@dlr.de',
    'maintainer': 'Thomas Baldauf',
    'maintainer_email': 'thomas.baldauf@dlr.de',
    'url': 'https://gitlab.dlr.de/bald_to/macro-pre-release',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
